# (a->b->b)->[a]->b
def foldr(f, slst, binit):
    while slst != []:
        last = slst[-1]
        slst = slst[:-1]
        binit = f(last, binit)
    return binit


# (a->b)->[a]->[b]
def lmap(f, aLst):
    return [f(a) for a in aLst]


# composition
def comp(f, g):
    return lambda x: f(g(x))


# string concat
def concat(sep):
    def myconcat(v1, v2):
        return str(v1) + sep + str(v2)
    return myconcat


# zip two list with function
# (a -> b -> c) -> [a] -> [b] -> [c]
def zipWith(f, alst, blst):
    res = []
    for i in range(min(len(alst), len(blst))):
        res.append(f(alst[i], blst[i]))
    return res


# zip three lists with function
# (a -> b -> c -> d) -> [a] -> [b] -> [c] -> [d]
def zipWith3(f, alst, blst, clst):
    res = []
    for i in range(min(len(alst), len(blst), len(clst))):
        res.append(f(alst[i], blst[i], clst[i]))
    return res


# filter
# (a -> bool) -> [a] -> [a]
def filter(f, aLst):
    return [a for a in aLst if f(a)]


# count number of items in list satisfy predicate
def countIf(p, lst):
    return len(filter(p, lst))




import numpy as np
import matplotlib.pyplot as plt
from .pyprelude import FPToolBox as fp
from .pyprelude.Progress import Progress

def plot1d(func, lb, ub, step, stop_val=None, stop_iter=50, svname='', show=False, esp=1e-7, adaptive=False):
    x = np.arange(lb, ub, step)
    prog = Progress(len(x), label='plotting', response_time=60)
    if adaptive:
        y = []
        stop_cnt = 0
        for xval in x:
            prog.count()
            yval = func(xval)
            y.append(yval)
            if abs(yval - stop_val) < esp:
                stop_cnt += 1
                if stop_cnt > stop_iter:
                    break
            else:
                stop_cnt = 0
        x = x[:len(y)]
    else:
        y = fp.lmap(func, x)

    fig = plt.figure(1)
    plt.minorticks_off()
    plt.plot(x, y, 'k', color='b')
    if svname != '':
        plt.savefig(svname)
    if show:
        plt.show()
    plt.close(fig)

#def find_ub(func, lb, step, lim, nochange_iter=6, eps=1e-6):
#    ub = lb
#    power = 0
#    y0 = func(ub)
#    while True:
#        ub += step * (2 ** power)
#        print(ub)
#        if ub > lim:
#            return ub
#
#        y1 = func(ub)
#        if abs(y0-y1) < eps:
#            if power > nochange_iter:
#                return ub
#            else:
#                y0 = y1
#                power += 1
#        else:
#            y0 = y1
#            power = 0
#


        




from . import Nregion as rg
from .numericFuncs import theta

def get_R(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val):
    # case e == f
    if e == f:
        if i == 0:
            a = 0
        else:
            a = d
        b = (d + le)/2
        if x_val is None:
            x_val = b
        xp = theta(a, b, x_val)
        if (i, j) == (0, 0):
            r = rg.RegionBase(0, 1, lambda p: max(0, p - xp/le), lambda p: p, x_val)
        elif (i, j) == (0, 1):
            r = rg.RegionBase(0, 1, lambda p: p, lambda p: min(1, p + xp/le), x_val)
        elif (i, j) == (1, 0):
            r = rg.RegionBase(1 - (xp - d)/le, 1, lambda p: 0, lambda p: p - 1 + (xp - d)/le, x_val)
        else:
            r = rg.RegionBase(0, (xp - d)/le,  lambda p: p + 1 - (xp - d)/le, lambda p: 1, x_val)
        r.a_val = a
        r.b_val = b
        return r
    else:
        a = d[i, j]
        b = (le + lf + 
                min(d[0, 0] + d[1, 1], 
                    d[0, 1] + d[1, 0])) / 2
        if x_val is None:
            x_val = b
        xp = theta(a, b, x_val)
        if (i, j) == (0, 0):
            r = rg.RegionBase(0, min(p1, (xp - d[i, j])/le), lambda p: 0, lambda p: min(q1, (xp - le * p - d[i, j])/lf), x_val, bd=(True, False))
        elif (i, j) == (0, 1):
            r = rg.RegionBase(0, min(p2, (xp - d[i, j])/le), lambda p: max(q1, (- xp + le * p + d[i, j] + lf)/lf), lambda p: 1, x_val, bd=(True, False))
        elif (i, j) == (1, 0):
            r = rg.RegionBase(max(p1, (- xp + d[i, j] + le)/le), 1, lambda p: 0, lambda p: min(q2, (xp + le * p - d[i, j] - le)/lf), x_val)
        else:
            r = rg.RegionBase(max(p2, (- xp + d[i, j] + le)/le), 1, lambda p: max(q2, (- xp - le * p + d[i, j] + le + lf)/lf), lambda p: 1, x_val)
        r.a_val = a
        r.b_val = b
        return r


# get L
def get_L(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val):
    if e == f:
        if i == 0:
            a = 0
        else:
            a = d
        b = (d + le)/2
        if x_val is None:
            x_val = b
        xp = theta(a, b, x_val)
        if (i, j) == (0, 0):
            r = rg.RegionBase(xp/le, 1, None, None, x_val)
        elif (i, j) == (0, 1):
            r = rg.RegionBase(0, 1 - xp/le, None, None, x_val)
        elif (i, j) == (1, 0):
            r = rg.RegionBase(1 - (xp - d)/le, 1, None, None, x_val)
        elif (i, j) == (1, 1):
            r = rg.RegionBase(0, (xp - d)/le, None, None, x_val)
        return r
    else:
        a = d[i, j]
        b = (le + lf + 
                min(d[0, 0] + d[1, 1], 
                    d[0, 1] + d[1, 0])) / 2
        if x_val is None:
            x_val = b
        xp = theta(a, b, x_val)
        if (i, j) == (0, 0):
            r = rg.RegionBase(
                    max(0,(2 * xp - d[0, 0] - d[0, 1] - lf)/(2 * le)), 
                    min(p1, (xp - d[i, j])/le), 
                    #Max(0,(2 * xp - d[0, 0] - d[0, 1] - lf)/(2 * le)), 
                    #Min(p1, (xp - d[i, j])/le), 
                    None, None, x_val, bd=(True, False))
        elif (i, j) == (0, 1):
            r = rg.RegionBase(
                    max(0, (2 * xp - d[0, 0] - d[0, 1] - lf)/(2 * le)), 
                    min(p2, (xp - d[i, j])/le), 
                    #Max(0, (2 * xp - d[0, 0] - d[0, 1] - lf)/(2 * le)), 
                    #Min(p2, (xp - d[i, j])/le), 
                    None, None, x_val, bd=(True, False))
        elif (i, j) == (1, 0):
            r = rg.RegionBase(
                    max(p1, (- xp + d[i, j] + le)/le), 
                    min(1, (- 2 * xp + d[1, 0] + d[1, 1] + 2* le + lf)/(2 * le)), 
                    #Max(p1, (- xp + d[i, j] + le)/le), 
                    #Min(1, (- 2 * xp + d[1, 0] + d[1, 1] + 2* le + lf)/(2 * le)), 
                    None, None, x_val)
        elif (i, j) == (1, 1):
            r = rg.RegionBase(
                    max(p2, (- xp + d[i, j] + le)/le), 
                    min(1, (- 2 * xp + d[1, 0] + d[1, 1] + 2* le + lf)/(2 * le)),
                    #Max(p2, (- xp + d[i, j] + le)/le), 
                    #Min(1, (- 2 * xp + d[1, 0] + d[1, 1] + 2* le + lf)/(2 * le)),
                    None, None, x_val)
        return r


def get_q(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val):
    x = x_val
    if e == f:
        if (i, j) == (0, 0):
            q_func = lambda p: p - x / le
        elif (i, j) == (0, 1):
            q_func = lambda p: p + x / le
        elif (i, j) == (1, 0):
            q_func = lambda p: p - 1 + (x - d) / le
        else:
            q_func = lambda p: p + 1 - (x - d) / le
        return q_func
    else:
        if (i, j) == (0, 0):
            q_func = lambda p: (x - le * p - d[i, j]) / lf
        elif (i, j) == (0, 1):
            q_func = lambda p: (- x + le * p + d[i, j] + lf) / lf
        elif (i, j) == (1, 0):
            q_func = lambda p: (x + le * p - d[i, j] - le) / lf
        else:
            q_func = lambda p: (- x - le * p + d[i, j] + le + lf) / lf
        return q_func



import time

class Progress:
    def __init__(self, total_iter, label='default', response_time = 60, on=True):
        self.label = label
        self.tot = total_iter * 1.0
        self.rt = response_time
        self.cnt = 0
        self.start = None
        self.on = on
        self.time0 = None

    def count(self):
        if self.on is False:
            return 0
        self.cnt += 1
        if self.start is None:
            self.time0 = time.time()
            self.start = self.time0
        else:
            end = time.time()
            if end - self.start > self.rt:
                print(self.label, 'progress:\t%0.4f%%\ttime:\t%0.1fs' % (self.cnt/self.tot * 100, end - self.time0))
                self.start = end







import time

class Timer:
    def __init__(self, name='timer', on=True):
        self.name = name
        self.startv = None
        self.endv = None
        self.t = -1
        self.tot = 0
        self.on = on

    def start(self):
        if self.on is False:
            return 0
        self.startv = time.time()

    def stop(self):
        if self.on is False:
            return 0
        self.endv = time.time()
        self.t = self.endv - self.startv
        self.tot += self.t

    def reset(self):
        if self.on is False:
            return 0
        self.t = -1
        self.startv = None
        self.endv = None
        self.tot = 0

    def show(self):
        if self.on is False:
            return 0
        print(self.name, ':\t', 'total ', self.tot, '\tlap ', self.t)






from . import FPToolBox as fp
from . import StringTools as st

# function that read data into double list
def dataFormat(ffname, start_row=0, seps='\t '):
    rf = open(ffname, 'r')
    lst = []
    for line in rf:
        line = line.strip()
        lst.append(fp.lmap(strip, st.splitOneOf(line, seps)))
    rf.close()
    return lst


# function for strip
def strip(strs):
    return strs.strip()



import networkx as nx
from .pyprelude import GraphGen as gg
from .commonFuncs import *
from sympy import *
import time


# reads a graph input file
def readGraph(fpath, gname):
    print('loading graph...')
    g = gg.GraphFromFile(fpath, gname).G
    g.name = gname
    return g

def basicInfo(g, phi, rational=False, d_jit=False):
    print('generating basic info...')
    g.rat = rational
    g.phi = phi

    # check graph info
    check_res = gcheck(g)
    if check_res['total'] is not True:
        raise Exception('Input did not pass all test: ', check_res)

    # dicts for global memorization
    g.dict_entry = {}

    # index set
    g.two = (0,1)
    g.two2 = [(i, j) for i in g.two for j in g.two]
    end = time.time()
    
    # get max edge length
    g.max_l = 0
    for e in g.edges():
        if g.edges[e]['l'] > g.max_l:
            g.max_l = g.edges[e]['l'] 

    g.d_max = -1
    g.dd_max = -1
    g.de_max = {}

    g.d = {}
    if d_jit is False:
        get_d(g)
        g.jit = False
    else:
        g.jit = True




# entry info
def entry_info(g, e, f, le, lf):
    p1 = None
    p2 = None
    q1 = None
    q2 = None
    if e == f:
        if g.jit is False:
            d = g.d[e[0]][e[1]]
        else:
            d = get_d_jit(g, e[0], e[1], e)
    else:
        d = {}
        if g.jit is False:
            for (i, j) in g.two2:
                d[i, j] = g.d[e[i]][f[j]]
        else:
            for (i, j) in g.two2:
                d[i, j] = get_d_jit(g, e[i], f[j], e)
        p1 = (-d[0, 0] + d[1, 0] + le) / (2 * le)
        p2 = (-d[0, 1] + d[1, 1] + le) / (2 * le)
        q1 = (-d[0, 0] + d[0, 1] + lf) / (2 * lf)
        q2 = (-d[1, 0] + d[1, 1] + lf) / (2 * lf)

    return(d, p1, p2, q1, q2)

# get d value just in time
def get_d_jit(g, u, v, e):
    if u in g.d:
        return g.d[u][v]
    elif v in g.d:
        return g.d[v][u]
    else:
        get_d(g, mye=e) 
        return g.d[u][v]


#########################################################################
# help functions                                                        #
#########################################################################

# get shortest path matrix D
def get_d(g, mye=None):
    if mye is None:
        print("loading all pairs shortest length...")
        g.d = dict(nx.shortest_path_length(g, weight = 'l'))
        update_dd_max(g)
    else:
        print("loading shortest length to nodes of edge", str(mye)+"...")
        g.d[mye[0]] = nx.shortest_path_length(g, source=mye[0], weight='l')
        g.d[mye[1]] = nx.shortest_path_length(g, source=mye[1], weight='l')
        update_dd_max(g, mye)

    # if g.rat:
    #     for i in g.nodes():
    #         for j in g.nodes():
    #             g.d[i][j] = rat(g.d[i][j], g.rat)

# update d_max
def update_dd_max(g, e=None):
    if e is None:
        if g.dd_max < 0:
            for i in g.nodes():
                for j in g.nodes():
                    if i < j:
                        if g.d[i][j] > g.dd_max:
                            g.dd_max = g.d[i][j]
            g.dd_max += g.max_l
    else:
        if e not in g.de_max:
            de_max = 0
            for j in g.d[e[0]]:
                if g.d[e[0]][j] > de_max:
                    de_max = g.d[e[0]][j]
            g.de_max[e] = de_max + g.edges[e]['l']









from sympy import sympify, lambdify, integrate
from sympy.abc import p, q
from scipy.integrate import dblquad

class Phi:
    def __init__(self, name, phi_pq=None, phi_p=None, phi_q=None, eps=1e-7):
        self.name = name
        self.phi_pq_S = phi_pq
        self.phi_p_S = phi_p
        self.phi_q_S = phi_q
        self.eps = eps
        self.__gen_phi()
        if self.__check() is False:
            raise Exception('input phi does not integrate to one')
        
    def __gen_phi(self):
        if self.phi_p_S is not None and self.phi_q_S is not None:
            self.phi_p_S = sympify(self.phi_p_S)
            self.phi_q_S = sympify(self.phi_q_S)
            self.phi_pq_S = self.phi_p_S * self.phi_q_S
            self.phi_qcp_S = self.phi_q_S
        elif self.phi_pq_S is not None:
            self.phi_pq_S = sympify(self.phi_pq_S)
            self.phi_p_S = integrate(self.phi_pq_S, (q, 0 , 1))
            self.phi_q_S = integrate(self.phi_pq_S, (p, 0 , 1))
            self.phi_qcp_S = self.phi_pq_S / self.phi_p_S
        else:
            raise Exception('please provide both phi_p and phi_q, or the joint pdf phi_pq.')

        self.phi_p_N = lambdify(p, self.phi_p_S)
        self.phi_q_N = lambdify(q, self.phi_q_S)
        self.phi_pq_N = lambdify((q, p), self.phi_pq_S)
        self.phi_qcp_N = lambdify((q, p), self.phi_qcp_S)

    def __check(self):
        phi_tot = dblquad(self.phi_pq_N, 0, 1, lambda p: 0, lambda p: 1)[0]
        if abs(phi_tot - 1) < self.eps:
            phi_one = True
        else:
            phi_one = False
        return phi_one




from .pyprelude import FPToolBox as fp
from .pyprelude import TableWriter as tw
from math import factorial
from sympy import *
from sympy.abc import a, b, x
import dill
import networkx as nx
from networkx import is_connected
from scipy.integrate import dblquad

# symbolic gate function
#eta = lambda a, b, x: Piecewise((1, (x >= a) & (x <= b)), (0, True))

class eta(Function):
    nargs = 3

    @classmethod
    def eval(cls, a, b, x):
        if a.is_Number and b.is_Number and x.is_Number:
            zero = numbers.Zero()
            if x >= a and x <= b:
                return 1 + zero
            else:
                return 0 + zero

class etal(Function):
    nargs = 3

    @classmethod
    def eval(cls, a, b, x):
        if a.is_Number and b.is_Number and x.is_Number:
            zero = numbers.Zero()
            if x > a and x <= b:
                return 1 + zero
            else:
                return 0 + zero

class etar(Function):
    nargs = 3

    @classmethod
    def eval(cls, a, b, x):
        if a.is_Number and b.is_Number and x.is_Number:
            zero = numbers.Zero()
            if x >= a and x < b:
                return 1 + zero
            else:
                return 0 + zero
# symbolic theshold function
#theta = lambda a, b, x: Piecewise((a, x < a), (b, x > b), (x, True))

class theta(Function):
    nargs = 3

    @classmethod
    def eval(cls, a, b, x):
        if a.is_Number and b.is_Number and x.is_Number:
            if x < a:
                return a
            elif x >b:
                return b
            else:
                return x

class mmin(Function):
    nargs = 2

    @classmethod
    def eval(cls, a, b):
        if a.is_Number and b.is_Number:
            if a <= b:
                return a
            else:
                return b

class mmax(Function):
    nargs = 2

    @classmethod
    def eval(cls, a, b):
        if a.is_Number and b.is_Number:
            if a >= b:
                return a
            else:
                return b

# return ordered set for B_curl(a,b)
# where a is the dimension of the tuple, and k is the sum
def B_curl(n, k):
    tmp = B_curl_h(n, k)
    return list(map(tuple, tmp))

def B_curl_h(n, k):
    if n == 1:
        return [[k]]

    if k == 0:
        return fp.lmap(lambda xs: addhd(0, xs), B_curl_h(n - 1, 0))
    
    # for each possible value, get one
    res = []
    for i in range(k + 1):
        res += fp.lmap(lambda xs: addhd(i, xs), B_curl_h(n - 1, k - i))

    return res

# the number of elements in B_curl
def B_curl_num(n, k):
    if n == 1:
        return 1

    if k == 0:
        return 1
    
    res = 0
    for i in range(k + 1):
        res += B_curl_num(n - 1, k - i)

    return res

# return ordered set for A_curl(a,b)
def A_curl(n, k):
    res = []
    for k0 in range(k + 1):
        res += B_curl(n, k0)

    return res

# return ordered set for A_curl(a,b)
def A_curl_num(n, k):
    res = 0
    for k0 in range(k + 1):
        res += B_curl_num(n, k0)

    return res

# function to fast add head to a list
def addhd(hd, xs):
    xs.insert(0, hd)
    return xs

#def ncr(n, r):
#    r = min(r, n-r)
#    numer = reduce(op.mul, range(n, n-r, -1), 1)
#    denom = reduce(op.mul, range(1, r+1), 1)
#    return numer / denom

# calc n choose multiindex rs

def ncrs(n, rs):
    res = factorial(n)
    r_sum = 0
    for r in rs:
        res = res / factorial(r)
        r_sum = r_sum + r

    return res / factorial(n - r_sum)

# search symetric tuples
#def get(mydict, key):
#    if key[0] > key[1]:
#        return mydict[key[1], key[0], key[2], key[3]]
#    else:
#        return mydict[key]


# rationalize input
def rat(val, rational):
    if rational is False:
        return float(val)
    else:
        return Rational(val)


if __name__ == '__main__':
    from sympy.plotting import plot
    
    plot(eta(1, oo, x), (x, -1, 5))

# load serialized formulas
# stat from enums
def load_formulas(gname, phiname, stat, folder='./.formulas/'):
    path = folder + gname + '_' + phiname + '_' + str(stat) + '.sav'
    dill.load(open(path, 'rb'))
    try:
        return dill.load(open(path, 'rb'))
    except:
        return None

# get largets components
def get_largest_component(g, save=True):
    if nx.is_connected(g):
        return g
    else:
        g = g.subgraph(max(nx.connected_components(g), key=len))
        prob_rescale(g, 'x')
        prob_rescale(g, 'y')
        if save:
            output = []
            head = ['i', 'j', 'l', 'x', 'y']
            output.append(head)
            for e in g.edges():
                erow = []
                e0 = min(int(e[0]), int(e[1]))
                e1 = max(int(e[0]), int(e[1]))
                erow.append(e0)
                erow.append(e1)
                erow.append(float(g.edges[e]['l']))
                erow.append(float(g.edges[e]['x']))
                erow.append(float(g.edges[e]['y']))
                output.append(erow)
            tw.tableWrite(output, './' + g.name + '_cc.dat', formatType='n')

        return g

# rescale prob
def prob_rescale(g, prop):
    tot = 0.0
    for e in g.edges():
        tot += g.edges[e][prop]
    for e in g.edges():
        g.edges[e][prop] = g.edges[e][prop] / tot

# check basic info of g
def gcheck(g, eps=1e-7):
    # is connected
    is_con = is_connected(g)
    # is p_x add up to 1
    px_tot = 0
    py_tot = 0
    # tolorence
    for e in g.edges():
        px_tot += g.edges[e]['x']
        py_tot += g.edges[e]['y']

    if abs(px_tot - 1) < eps:
        px_one = True
    else:
        px_one = False

    if abs(py_tot - 1) < eps:
        py_one = True
    else:
        py_one = False

    return {'connected': is_con, 
            'px_one': (px_one, float(px_tot)), 
            'py_one': (py_one, float(py_tot)),
            'total': (is_con and px_one and py_one)}







from scipy.integrate import quad, dblquad
from sympy import lambdify

# numeric gate function
def eta(a, b, x):
    if x >= a and x <= b:
        return 1
    else:
        return 0

# numeric gate function
def etal(a, b, x):
    if x > a and x <= b:
        return 1
    else:
        return 0

# numeric gate function
def etar(a, b, x):
    if x >= a and x < b:
        return 1
    else:
        return 0

# numeric threshold function
def theta(a, b, x):
    if x < a:
        return a
    elif x >b:
        return b
    else:
        return x

# numerical integration with an input expression from sympy
def N_int(expr, range1, range2=None):
    mods = ['numpy', {'theta': theta, 'eta': eta, 'etal': etal, 'etar': etar, 'mmin': min, 'mmax': max}]

    x = range1[0]
    region = [range1[1], range1[2]]

    if range2 is not None:
        y = range2[0]
        var = (y, x)
        yl = lambdify(x, range2[1], modules=mods)
        yu = lambdify(x, range2[2], modules=mods)
        region.append(yl)
        region.append(yu)
        int_func = dblquad
    else:
        var = x
        int_func = quad

    region = tuple(region)

    f = lambdify(var, expr, modules=mods)

    return int_func(f, *region)

# check if input expression is a valid pdf 
# can only verify the integral as 1
# the non-neg need to be guarenteed manually
# only support at most 2-dim input function
def pdf_check(expr, range1, range2=None):
    m, e = N_int(expr, range1, range2)

    if 1 < m - e or 1 > m + e:
        return (False, m, e)
    else:
        return (True, m, e)

    
if __name__ == '__main__':
    from sympy.abc import x,y
    expr = 4 * x * y
    range1 = (x, 0, 1)
    range2 = (y, 0, 1)
    print(pdf_check(expr, range1, range2))








def memorize(ddict, key, func, param_lst=None, param_dict=None, on=True):
    if on is False:
        return runfunc(func, param_lst, param_dict)
    else:
        if key not in ddict:
            ddict[key] = runfunc(func, param_lst, param_dict)

        return ddict[key]


def runfunc(func, param_lst, param_dict):
    if param_lst is None and param_dict is None:
        return func()
    elif param_lst is not None and param_dict is None:
        return func(*param_lst)
    elif param_lst is None and param_dict is not None:
        return func(**param_dict)
    else:
        return func(*param_lst, **param_dict)






from .formulas import Formulas
from .enums import Stats

# info_dict keys and values
# moment: ks; cdf: nothing; pdf: nothing;
# cmoment: (ks, locs); ccdf: locs; cpdf: locs
def data_collector(gname, phi, mmtp=False, cdfp=False, pdfp=False, cmmtp=False, ccdfp=False, cpdfp=False, d_jit=False, memorize=True):
    fls = Formulas(gname, phi, d_jit=d_jit, memorize=memorize)

    # moments
    if mmtp is not False and mmtp['collect']:
        moment = fls.get_formula(Stats.MOMENT, symbolic=mmtp['symbolic'])
        for k in mmtp['valst']:
            moment.eval(k)

    # cdf
    if cdfp is not False and cdfp['collect']:
        cdf = fls.get_formula(Stats.CDF, symbolic=cdfp['symbolic'])
        cdf.plot()
        if cdfp['symbolic']:
            cdf.save_formulas()

    # pdf
    if pdfp is not False and pdfp['collect']:
        pdf = fls.get_formula(Stats.PDF, symbolic=pdfp['symbolic'])
        pdf.plot()
        if pdfp['symbolic']:
            pdf.save_formulas()

    # cmoments
    if cmmtp is not False and cmmtp['collect']:
        cmoment = fls.get_formula(Stats.CMOMENT, symbolic=cmmtp['symbolic'])
        for k in cmmtp['valst'][0]:
            for loc in cmmtp['valst'][1]:
                cmoment.eval(k, *loc)

    # ccdf
    if ccdfp is not False and ccdfp['collect']:
        ccdf = fls.get_formula(Stats.CCDF, symbolic=ccdfp['symbolic'])
        for loc in ccdfp['valst']:
            ccdf.plot(*loc)
        if ccdfp['symbolic']:
            ccdf.save_formulas()

    # cpdf
    if cpdfp is not False and cpdfp['collect']:
        cpdf = fls.get_formula(Stats.CPDF, symbolic=cpdfp['symbolic'])
        for loc in cpdfp['valst']:
            cpdf.plot(*loc)
        if cpdfp['symbolic']:
            cpdf.save_formulas()




from . import FPToolBox as fp

# split one of the char in
def splitOneOf(s, seps, delEmpty=True):
    sep0 = seps[0]
    sepTail = seps[1:]
    sLst = s.split(sep0)
    for c in sepTail:
        tmpLst = fp.lmap(lambda x: x.split(c), sLst)
        sLst = fp.foldr(lambda x, y: x + y, tmpLst, [])

    if delEmpty is True:
        sLst = fp.filter(lambda x: x != '', sLst)

    return sLst




from . import FPToolBox as fp
from . import EasyWriter as ew

""" output a table from nested list input"""
def tableStrFormat(lst, cSep='\t', rSep='\n'):

    cConcat = fp.concat(cSep)
    rConcat = fp.concat(rSep)

    def formRow(lst1):
        return fp.foldr(cConcat, fp.lmap(str, lst1), '')[:-1]

    return fp.foldr(rConcat, fp.lmap(formRow, lst), '')[:-1]


""" form and write the table """
# formatType: b: bueaty, n: normal
def tableWrite(lst, wfname, cSep='\t', rSep='\n', wtype='w', formatType='b', numCols=[], padNum=2):
    if formatType == 'b':
        output = tableBeautyFormat(lst, numCols, padNum)
    else:
        output = tableStrFormat(lst, cSep, rSep)
    ew.wFile(wfname, output, wtype)


""" formart list of lists with better presentation """
def tableBeautyFormat(lst0, numCols=[], padNum=2):

    # change all value to string
    lst = []
    for row in lst0:
        lst.append(fp.lmap(str, row))

    # get the nth column
    maxLen = []
    for i in range(len(lst[0])):
        tmpCol = nthCol(lst, i)
        lenLst = fp.lmap(len, tmpCol)
        maxLen.append(max(lenLst) + padNum)
    
    # direction list
    dirLst = ['left'] * len(lst[0])
    for i in numCols:
        dirLst[i] = 'right'
    
    # for each row do padding
    def resize(vlst, lenLst, dirLst):
        res = fp.zipWith3(padding, vlst, maxLen, dirLst)
        return res

    bueatyLst = fp.lmap(lambda x: resize(x, lenLst, dirLst), lst)
    return tableStrFormat(bueatyLst, cSep='')


""" padding function"""
def padding(s, padLen, dirc, padChar=' '):
    if padLen <= len(s):
        return s
    l = padLen - len(s)
    if dirc == 'left':
        s1 = padChar + s + (padChar * l)
    else:
        s1 = (padChar * l) + s + padChar
    return s1


""" get the nth column of a nested list """
def nthCol(lst, n, f=lambda x: x):
    return fp.lmap(lambda x: f(x[n]), lst)
    


# mylst = [["testa", 2, 3], [2, "testttttb", 4], [4, 5, "testc01"]]
# print(tableBeautyFormat(mylst, [1]))
# print(padding('testa', 7, 'left', padChar=' '))
# print(tableStrFormat(mylst))



from . import commonFuncs as cf
from . import numericFuncs as nf
from . import basicInfo as bi
from .pyprelude.Timer import Timer
from .pyprelude.Progress import Progress
from .pyprelude import EasyWriter as ew
from .enums import Stats
from sympy import symbols, lambdify
from sympy.abc import x, p, q
from .serializedFormula import *
from .plot import plot1d
from .pyprelude.Memorization import memorize
import os
import dill


# template for all region-wise operations
class Formula:
    def __init__(self, g, memorize):
        self.g = g
        self.stat = None
        self.keys = None
        self.unikey = None
        self.idx_num = 0
        self.val_keys = None
        self.resfolder = './results/'
        self.plot_info = [None, x, -0.1, None]  # plotting parameters
        self.memo = memorize
        self.memo_dict()
        self.mods = ['numpy', {'theta': nf.theta, 'eta': nf.eta, 'etal': nf.etal, 'etar': nf.etar, 'mmin': min, 'mmax': max}]  ## modules for lambdify

    # set up the symbolic or numeric environment
    def sym_num_env(self, symbolic):
        self.symbolic = symbolic
        g = self.g
        if symbolic:
            # assgin symbolic functions
            from .Sregions import get_R, get_L, get_q
            from .commonFuncs import eta, etal, etar
            self.get_R = get_R
            self.get_L = get_L
            self.get_q = get_q
            self.eta = eta
            self.etal = etal
            self.etar = etar
            
            # assign symbolic phis
            g.phi_p = g.phi.phi_p_S
            g.phi_q = g.phi.phi_q_S
            g.phi_pq = g.phi.phi_pq_S
            g.phi_qcp = g.phi.phi_qcp_S

            # set extra properties
            self.fs = {} # all formulas
            self.N_fs = {} # all numeric formulas
            self.unikey = 'zero'
            self.folder = './.formulas/'
            self.mksvname()

        else:
            # assgin symbolic functions
            from .Nregions import get_R, get_L, get_q
            from .numericFuncs import eta, etal, etar
            self.get_R = get_R
            self.get_L = get_L
            self.get_q = get_q
            self.eta = eta
            self.etal = etal
            self.etar = etar
            
            # assign symbolic phis
            g.phi_p = g.phi.phi_p_N
            g.phi_q = g.phi.phi_q_N
            g.phi_pq = g.phi.phi_pq_N
            g.phi_qcp = g.phi.phi_qcp_N

    def memo_dict(self):
        self.dict_coeff = {}
        self.dict_prob = {}
        self.dict_ef_entry = {}

    def cond_ep(self, g, k, e, p_val, x_val, alphas, prog):
        res = 0
        # get e related info
        le = float(g.edges[e]['l'])
        px = float(g.edges[e]['x'])
        for f in g.edges():

            prog.count()

            # get f related info
            lf = float(g.edges[f]['l'])
            py = float(g.edges[f]['y'])

            # get entry coeff
            prob = memorize(self.dict_prob, (e, f), self.ef_prob, (px, py), on=self.memo)

            ef_res = self.ef_entry_val_help(g, k , e, f, p_val, x_val, alphas, le, lf)
#            # get entry ef related info
#            d, p1, p2, q1, q2 = memorize(self.g.dict_entry, (e, f), bi.entry_info,(self.g, e, f, le, lf), on=self.memo)
#            #d, p1, p2, q1, q2 = bi.entry_info(self.g, e, f, le, lf)
#
#            # get entry coeff
#            coeff1, coeff2 = memorize(self.dict_coeff, (e, f), self.ef_coeff, (px, py, lf), on=self.memo)
#
#            ef_res = 0
#
#            for (i, j) in g.two2:
#                # get region related info
#                R = self.get_R(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val)
#                ef_res += self.region_op(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, R, p_val, x_val, k, alphas)

            res += prob * ef_res
        return res


    def ef_entry_val_help(self, g, k, e, f, p_val, x_val, alphas, le, lf):

        # get entry ef related info
        d, p1, p2, q1, q2 = memorize(self.g.dict_entry, (e, f), bi.entry_info,(self.g, e, f, le, lf), on=self.memo)

        # get entry coeff
        coeff = memorize(self.dict_coeff, (e, f), self.ef_coeff, (lf,), on=self.memo)

        ef_res = 0

        for (i, j) in g.two2:
            # get region related info
            R = self.get_R(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val)
            ef_res += self.region_op(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, R, p_val, x_val, k, alphas)

        return coeff * ef_res


    def ef_entry_val(self, e, f, k=None, p_val=None, x_val=None):
        g = self.g
        # generate extra params
        alphas = None
        if self.stat == Stats.MOMENT or self.stat == Stats.CMOMENT:
            alphas = cf.A_curl(2, k)

        # get e related info
        le = float(g.edges[e]['l'])

        # get f related info
        lf = float(g.edges[f]['l'])

        ef_res = self.ef_entry_val_help(self.g, k , e, f, p_val, x_val, alphas, le, lf)

        return ef_res

    def ef_entry_memo(self, e, f, k_val=None, p_val=None, x_val=None):
        return memorize(self.dict_ef_entry, (e, f, k_val, p_val, x_val), self.ef_entry_val, (e, f, k_val, p_val, x_val), on=self.memo)

    def valuefy(self, expr, symbols, vals):
        if 'sympy' in str(type(expr)):
            myf = lambdify(symbols, expr, modules=self.mods)
            return myf(*vals)
        else:
            return expr

    def X_coeff(self, k_val=None, p_val=None, x_val=None):
        g = self.g
        coeff = {}
        for e in self.g.edges():
            val = 0
            for f in self.g.edges():
                py = float(g.edges[f]['y'])
                val += py * self.ef_entry_val(e, f, k_val, p_val, x_val)
                #val += py * self.ef_entry_memo(e, f, k_val, p_val, x_val)
            coeff[e] = self.valuefy(val, (p, x), (p_val, x_val))
        return coeff

    def Y_coeff(self, k_val=None, p_val=None, x_val=None):
        g = self.g
        coeff = {}
        for f in self.g.edges():
            val = 0
            for e in self.g.edges():
                px = float(g.edges[e]['x'])
                val += px * self.ef_entry_val(e, f, k_val, p_val, x_val)
                #val += px * self.ef_entry_memo(e, f, k_val, p_val, x_val)
            coeff[f] = self.valuefy(val, (p, x), (p_val, x_val))
        return coeff

    def Y_coeff_condi(self, e, p_val, k_val=None, x_val=None):
        g = self.g
        coeff = {}
        for f in self.g.edges():
            val = self.ef_entry_val(e, f, k_val, p_val, x_val)
            coeff[f] = self.valuefy(val, (x,), (x_val,))
            #coeff[f] = self.ef_entry_memo(e, f, k_val, p_val, x_val)
        return coeff

    def cond_region(self, k=None, e=None, p_val=None, x_val=None, show=True):
        # generate extra params
        alphas = None
        if self.stat == Stats.MOMENT or self.stat == Stats.CMOMENT:
            alphas = cf.A_curl(2, k)

        g = self.g
        res = 0

        if show:
            print('computing', g.name, g.phi.name, str(self.stat)+'...')


        # treat conditional
        if e is not None:
            # see progress
            prog = Progress(g.number_of_edges(), label='computing', response_time=60)
            return self.cond_ep(g, k, e, p_val, x_val, alphas, prog)

        # see progress
        prog = Progress(g.number_of_edges() ** 2, response_time=60)

        for e in g.edges():
            res += self.cond_ep(g, k, e, p_val, x_val, alphas, prog)
        return res
    
    # combine results in e,f level
    def ef_coeff(self, lf):
        if self.stat == Stats.MOMENT:
            coeff = 1
        elif self.stat == Stats.CDF:
            coeff = 1
        elif self.stat == Stats.PDF:
            coeff = 1 / lf
        elif self.stat == Stats.CMOMENT:
            coeff = 1
        elif self.stat == Stats.CCDF:
            coeff = 1
        elif self.stat == Stats.CPDF:
            coeff = 1 / lf
        return coeff

    # combine results in e,f level
    def ef_prob(self, px, py):
        if self.stat == Stats.MOMENT:
            coeff = px * py
        elif self.stat == Stats.CDF:
            coeff = px * py
        elif self.stat == Stats.PDF:
            coeff = px * py
        elif self.stat == Stats.CMOMENT:
            coeff = py
        elif self.stat == Stats.CCDF:
            coeff = py
        elif self.stat == Stats.CPDF:
            coeff = py
        return coeff

    # save result when evaluate
    def save_res(self, keys, res):
        g = self.g
        fname = self.resfolder + g.name + '_' + g.phi.name + '_' + str(self.stat) + '.dat'
        val = str(keys) + '\t' + str(res) + '\n'
        ew.wFile(fname, val)

    # make save name
    def mksvname(self):
        self.fullsave = self.folder + self.g.name + '_' + self.g.phi.name + '_' + str(self.stat) + '.sav'

    # make dict from input
    def make_params(self, *params):
        p_dict = {}
        val_dict = {}
        keys = []
        vals = []
        for idx, v in enumerate(params):
            if idx < len(self.keys):
                if self.keys[idx] == 'p_val':
                    # use limits instead of undefined p value
                    v = self.adj_p_val(v)
                    val_dict['p_val'] = v
                    vals.append(v)
                    if self.symbolic:
                        p_dict['p_val'] = p
                    else:
                        p_dict['p_val'] = v
                elif self.keys[idx] == 'x_val':
                    val_dict['x_val'] = v
                    vals.append(v)
                    if self.symbolic:
                        p_dict['x_val'] = x
                    else:
                        p_dict['x_val'] = v
                else:
                    keys.append(v)
                    p_dict[self.keys[idx]] = v
            else:
                break

        if keys == []:
            keys = [self.unikey]

        keys = tuple(keys)
        vals = tuple(vals)

        return p_dict, val_dict, keys, vals

    # use limit to replace undefined p
    def adj_p_val(self, p_val, eps=1e-7):
        if p_val > 0 and p_val < 1:
            return p_val
        elif p_val == 0:
            return eps
        elif p_val == 1:
            return 1 - eps
        else:
            raise Exception('p_val has to be in [0, 1]')

    ### abstract methods
    # operations in each region
    def region_op(self, g, e, f, i, j, le, lf, px, py, d, p1, p2, q1, q2, R, p_val, x_val, k, alphas):
        pass


class Moment(Formula):
    def __init__(self, g, symbolic, memorize):
        super().__init__(g, memorize)
        self.moment_info = {}
        self.symbolic = symbolic
        self.stat = Stats.MOMENT
        self.idx_num = 1
        self.keys = ['k']
        self.sym_num_env(symbolic)

    def get_func(self, g, alpha):
        if self.symbolic:
            p, q = symbols('p,q')
            func = q ** alpha[0] * p ** alpha[1] * g.phi_pq
        else:
            func = lambda q, p: q ** alpha[0] * p ** alpha[1] * g.phi_pq(q, p)
        return func

    def get_const(self, g, e, f, i, j, le, lf, R, alpha):
        c0 = lf ** alpha[0] * le ** alpha[1]
        if e == f:
            w = (i + j + 1) * alpha[0] + (i + j + 2) * alpha[1]
            c1 = (-1) ** w 
        else:
            w = j * alpha[0] + i * alpha[1]
            c1 = (-1) ** w 
        c = c0 * c1
        func = self.get_func(g, alpha)

        m = R.m(func)
        return c * m


    def region_op(self, g, e, f, i, j, le, lf, d, p1, p2, q1, q2, R, p_val, x_val, k, alphas):
        res = 0
        # get alpha related info
        for alpha in alphas:
            const = memorize(
                    self.moment_info, 
                    (e, f, i, j, alpha), 
                    self.get_const, 
                    (g, e, f, i, j, le, lf, R, alpha),
                    on=self.memo
                    )

            if e == f:
                c2 = cf.ncrs(k, alpha) * (i * (d + le)) ** (k - alpha[0] - alpha[1])
            else:
                c2 = cf.ncrs(k, alpha) * (d[i, j] + i * le + j * lf) ** (k - alpha[0] - alpha[1])

            res += c2 * const
        return res


class CDF(Formula):
    def __init__(self, g, symbolic, memorize):
        super().__init__(g, memorize)
        self.stat = Stats.CDF
        self.symbolic = symbolic
        self.keys = ['x_val']
        self.val_keys = (x,)
        self.sym_num_env(symbolic)
        self.stop_val = 1

    def region_op(self, g, e, f, i, j, le, lf, d, p1, p2, q1, q2, R, p_val, x_val, k, alphas):
        return R.m(g.phi_pq)


class PDF(Formula):
    def __init__(self, g, symbolic, memorize):
        super().__init__(g, memorize)
        self.stat = Stats.PDF
        self.symbolic = symbolic
        self.keys = ['x_val']
        self.val_keys = (x,)
        self.sym_num_env(symbolic)
        self.stop_val = 0

    def get_func(self, q_func, g):
        if self.symbolic:
            p, q = symbols('p,q')
            func = g.phi_pq.subs(q, q_func)
        else:
            func = lambda p: g.phi_pq(q_func(p), p)
        return func

    def region_op(self, g, e, f, i, j, le, lf, d, p1, p2, q1, q2, R, p_val, x_val, k, alphas):
        c = self.eta(R.a_val, R.b_val, x_val)
        L = self.get_L(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val)
        q_func = self.get_q(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val)
        phis = self.get_func(q_func, g)
        m = L.m(phis)
        return c * m


class CMoment(Formula):
    def __init__(self, g, symbolic, memorize):
        super().__init__(g, memorize)
        self.stat = Stats.CMOMENT
        self.symbolic = symbolic
        self.idx_num = 2
        self.keys = ['k', 'e', 'p_val']
        self.val_keys = (p,)
        self.sym_num_env(symbolic)
        self.plot_info = [None, p, 0, 1]
        self.stop_val = 0

    def get_func(self, g, p_val, alpha):
        if self.symbolic:
            p, q = symbols('p,q')
            func = q ** alpha[0] * g.phi_qcp
        else:
            func = lambda q: q ** alpha[0] * g.phi_qcp(q, p_val)
        return func

    def region_op(self, g, e, f, i, j, le, lf, d, p1, p2, q1, q2, R, p_val, x_val, k, alphas):
        res = 0
        for alpha in alphas:
            c0 = cf.ncrs(k, alpha) * lf ** alpha[0] * le ** alpha[1]
            if e == f:
                w = (i + j + 1) * alpha[0] + (i + j + 2) * alpha[1]
                c1 = (-1) ** w * (i * (d + le)) ** (k - alpha[0] - alpha[1])
            else:
                w = j * alpha[0] + i * alpha[1]
                c1 = (-1) ** w * (d[i, j] + i * le + j * lf) ** (k - alpha[0] - alpha[1])
            c = c0 * c1 * p_val ** alpha[1]

            func = self.get_func(g, p_val, alpha)

            m = R.m_p(func, p_val)
            res += c * m

        return res


class CCDF(Formula):
    def __init__(self, g, symbolic, memorize):
        super().__init__(g, memorize)
        self.stat = Stats.CCDF
        self.symbolic = symbolic
        self.idx_num = 1
        self.keys = ['e', 'p_val', 'x_val']
        self.val_keys = (p, x)
        self.sym_num_env(symbolic)
        self.stop_val = 1

    def get_func(self, g, p_val):
        if self.symbolic:
            p, q = symbols('p,q')
            func = g.phi_qcp
        else:
            func = lambda q: g.phi_qcp(q, p_val)
        return func

    def region_op(self, g, e, f, i, j, le, lf, d, p1, p2, q1, q2, R, p_val, x_val, k, alphas):
        func = self.get_func(g, p_val)
        return R.m_p(func, p_val)


class CPDF(Formula):
    def __init__(self, g, symbolic, memorize):
        super().__init__(g, memorize)
        self.stat = Stats.CPDF
        self.symbolic = symbolic
        self.idx_num = 1
        self.keys = ['e', 'p_val', 'x_val']
        self.val_keys = (p, x)
        self.sym_num_env(symbolic)
        self.stop_val = 0

    def get_func(self, g, q_func, p_val):
        if self.symbolic:
            func = g.phi_qcp.subs(q, q_func)
        else:
            func = g.phi_qcp(q_func(p_val), p_val)
        return func

    def region_op(self, g, e, f, i, j, le, lf, d, p1, p2, q1, q2, R, p_val, x_val, k, alphas):
        c0 = self.eta(R.a_val, R.b_val, x_val)
        L = self.get_L(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val)
        q_func = self.get_q(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val)
        myeta = None
        if L.bd[0] is False:
            myeta = self.etal
        elif L.bd[1] is False:
            myeta = self.etar
        else:
            myeta = self.eta
        c1 = myeta(L.pl, L.pu, p_val)
        phis = self.get_func(g, q_func, p_val)
        return c0 * c1 * phis


# the interface for numerical computation
class Numeric:
    # fl_cls is the formula class, e.g. Moment, CDF, CMoment, etc.
    def __init__(self, g, fl_cls, memorize=False):
        # gen formula
        self.fm = fl_cls(g, False, memorize=memorize)

    # return coeff
    def X_coeff(self, k_val=None, p_val=None, x_val=None):
        return self.fm.X_coeff(k_val, p_val, x_val)

    # return coeff
    def Y_coeff(self, k_val=None, p_val=None, x_val=None):
        return self.fm.Y_coeff(k_val, p_val, x_val)

    # return coeff
    def Y_coeff_condi(self, e, p_val, k_val=None, x_val=None):
        return self.fm.Y_coeff(k_val, p_val, x_val)

    # evaluate value
    def eval(self, *params, save=True):
        fm = self.fm
        p_dict, v_dict, keys, vals = fm.make_params(*params)
        res = fm.cond_region(**p_dict)
        if save:
            fm.save_res(params, res)
        return res

    # numerical plot
    def plot(self, *params, pnts=1000, save=True, show=False):
        fm = self.fm
        if fm.stat == Stats.MOMENT:
            return 0
        p_dict, v_dict, keys, vals = fm.make_params(*params)
        x_key = fm.keys[len(fm.keys)-1]

        def f_lambda(x):
            p_dict[x_key] = x
            return fm.cond_region(show=False, **p_dict)
        
        var = fm.plot_info[1]
        lb = fm.plot_info[2]
        ub = fm.plot_info[3]

        # get upper bound
        if ub is None:
            print('updating plotting bounds...')
            if fm.g.d_max > 0:
                ub = fm.g.d_max
                adaptive = False
            elif fm.g.jit and 'e' in fm.keys:
                tmpe = p_dict['e']
                bi.get_d_jit(fm.g, tmpe[0], tmpe[1], tmpe)
                ub = fm.g.de_max[tmpe]
                adaptive = True
            else:
                ub = fm.g.dd_max
                adaptive = True
        else:
            adaptive = False

        step = ub / float(pnts)
        g = fm.g
        print('plotting', g.name, g.phi.name, str(fm.stat)+'...')

        if save:
            g = fm.g
            figname = fm.resfolder + g.name + '_' + g.phi.name + '_' + str(fm.stat) + '_' + str(params) + '.png'
            plot1d(f_lambda, lb, ub, step, svname=figname, show=show, stop_val=fm.stop_val, adaptive=adaptive)
        else:
            plot1d(f_lambda, lb, ub, step, show=show, stop_val=fm.stop_val, adaptive=adaptive)


# the interface for symbolic computation
class Symbolic:
    # fl_cls is the formula class, e.g. Moment, CDF, CMoment, etc.
    def __init__(self, g, fl_cls, memorize=False):

        # gen formula
        self.fm = fl_cls(g, True, memorize=memorize)

    # generate functions
    def gen_formula(self, *params):
        fm = self.fm
        p_dict, v_dict, keys, vals = fm.make_params(*params)
        p_dict = self.add_default_var(p_dict)
        if keys not in fm.fs:
            expr = fm.cond_region(**p_dict)
            if fm.plot_info[3] is None:
                fm.plot_info[3] = fm.g.d_max + 1
            fm.fs[keys] = expr
            if fm.val_keys is None:
                fm.N_fs[keys] = expr
            else:
                fm.N_fs[keys] = lambdify(fm.val_keys, expr, modules=fm.mods)
        return v_dict, keys, vals, fm.fs[keys], fm.N_fs[keys]

    # add default
    def add_default_var(self, p_dict):
        fm = self.fm
        if 'x_val' in fm.keys and 'x_val' not in p_dict:
            p_dict['x_val'] = x
        if 'p_val' in fm.keys and 'p_val' not in p_dict:
            p_dict['p_val'] = p
        return p_dict

    # plot formulas
    def plot(self, *params, step=0.01, save=True, show=False):
        fm = self.fm
        if fm.stat == Stats.MOMENT:
            return 0

        v_dict, keys, vals, f, N_f = self.gen_formula(*params)
        my_f = self.partial_apply(v_dict, f)

        var = fm.plot_info[1]
        lb = fm.plot_info[2]
        ub = fm.plot_info[3]

        f_lambda = lambdify(var, my_f, modules=fm.mods)

        if save:
            g = fm.g
            figname = fm.resfolder + g.name + '_' + g.phi.name + '_' + str(fm.stat) + '_' + str(params) + '.png'
            plot1d(f_lambda, lb, ub, step, svname=figname, show=show)
        else:
            plot1d(f_lambda, lb, ub, step, show=show)

    # evaluate value
    def eval(self, *params, save=True):
        fm = self.fm
        v_dict, keys, vals, f, N_f = self.gen_formula(*params)
        if len(vals) == 0:
            res = N_f
        else:
            res = N_f(*vals)
        if save:
            fm.save_res(params, res)
        return res

    # return coeff
    def X_coeff(self, k_val=None, p_val=None, x_val=None):
        return self.fm.X_coeff(k_val, p_val, x_val)

    # return coeff
    def Y_coeff(self, k_val=None, p_val=None, x_val=None):
        return self.fm.Y_coeff(k_val, p_val, x_val)

    # return coeff
    def Y_coeff_condi(self, e, p_val, k_val=None, x_val=None):
        return self.fm.Y_coeff(k_val, p_val, x_val)

    # make sublist
    def make_sub_lst(self, v_dict):
        subval = []
        if 'x_val' in v_dict:
            subval.append((x, v_dict['x_val']))

        if 'p_val' in v_dict:
            subval.append((p, v_dict['p_val']))
        return subval

    # print formula
    def formula(self, *params):
        fm = self.fm
        v_dict, keys, vals, f, N_f = self.gen_formula(*params)
        return self.partial_apply(v_dict, f)

    # partial apply
    def partial_apply(self, v_dict, f):
        subval = self.make_sub_lst(v_dict)
        if len(subval) == 0:
            return f
        else:
            return f.subs(subval)

    # save formulas
    def save_formulas(self):
        fm = self.fm
        sf = SFormula()
        sf.stat = fm.stat
        sf.fs = fm.fs
        sf.N_fs = fm.N_fs
        sf.mods = fm.mods
        sf.idx_num = fm.idx_num
        sf.val_keys = fm.val_keys
        sf.plot_info = fm.plot_info
        sf.unikey = fm.unikey

        # create folder
        if not os.path.exists(fm.folder):
            os.makedirs(fm.folder)

        # pickle sf
        dill.settings['recurse'] = True
        dill.dump(sf, open(fm.fullsave, 'wb'))

        return sf

# the wrapper for getting formulas
class Formulas:
    # symbolic: use symbolic or numerical method, None is same as auto
    def __init__(self, gname, phi, fpath='./data/', rational=False, d_jit=False, memorize=True):
        self.d_jit = d_jit
        self.memo = memorize
        # read graph
        self.g = bi.readGraph(fpath, gname)
        # generate basic info
        bi.basicInfo(self.g, phi, rational, d_jit)
        # dispatch map
        self.fl = {
                Stats.MOMENT: Moment,
                Stats.CDF: CDF,
                Stats.PDF: PDF,
                Stats.CMOMENT: CMoment,
                Stats.CCDF: CCDF,
                Stats.CPDF: CPDF
                }
        # conditional names
        self.condi = [
                Stats.CMOMENT,
                Stats.CCDF,
                Stats.CPDF
                ]

    # get different stats
    def get_formula(self, stats, symbolic=None):
        g = self.g
        if Stats.is_member(stats) is False:
            raise Exception('stats must be value from the enum Stats')

        if symbolic is None:
            if stats == Stats.MOMENT or stats == Stats.CMOMENT:
                symbolic = False
            else:
                symbolic = True

        if symbolic:
            return Symbolic(g, self.fl[stats], self.memo)
        else:
            return Numeric(g, self.fl[stats], self.memo)



from . import region as rg
from sympy import *
from sympy.abc import x, p, q
from .commonFuncs import theta, mmax, mmin

def get_R(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val):
    # case e == f
    if e == f:
        if i == 0:
            a = 0
        else:
            a = d
        b = (d + le)/2
        if x_val is None:
            x_val = b
        xp = theta(a, b, x_val)
        if (i, j) == (0, 0):
            r1 = rg.RegionBase(0, xp/le, 0, p, x_val, bd=(True, False))
            r2 = rg.RegionBase(xp/le, 1, p - xp/le, p, x_val)
            R = rg.Region((r1,r2))
        elif (i, j) == (0, 1):
            r1 = rg.RegionBase(0, 1 - xp/le, p, p + xp/le, x_val, bd=(True, False))
            r2 = rg.RegionBase(1 - xp/le, 1, p, 1, x_val)
            R = rg.Region((r1,r2))
        elif (i, j) == (1, 0):
            r1 = rg.RegionBase(1 - (xp - d)/le, 1, 0, p - 1 + (xp - d)/le, x_val)
            R = rg.Region((r1,))
        else:
            r1 = rg.RegionBase(0, (xp - d)/le,  p + 1 - (xp - d)/le, 1, x_val)
            R = rg.Region((r1,))
        R.a_val = a
        R.b_val = b
        if b > g.d_max:
            g.d_max = b
        return R
    else:
        a = d[i, j]
        b = (le + lf + 
                min(d[0, 0] + d[1, 1], 
                    d[0, 1] + d[1, 0])) / 2
        if x_val is None:
            x_val = b
        xp = theta(a, b, x_val)
        if (i, j) == (0, 0):
            r1 = rg.RegionBase(0, (xp - d[i, j])/le, 0, (xp - le * p - d[i, j])/lf, x_val)
            r2 = rg.RegionBase(p1, (xp - d[i, j])/le, 0, (xp - le * p - d[i, j])/lf, x_val, sgn=-1, eta=True, a=le * p1 + d[i, j])
            r3 = rg.RegionBase(0, (xp - lf * q1 - d[i, j])/le, q1, (xp - le * p - d[i, j])/lf, x_val, sgn=-1, eta=True, a=lf * q1 + d[i, j])
            R = rg.Region((r1, r2, r3))
        elif (i, j) == (0, 1):
            r1 = rg.RegionBase(0, (xp - d[i, j])/le, (- xp + le * p + d[i, j] + lf)/lf, 1, x_val)
            r2 = rg.RegionBase(p2, (xp - d[i, j])/le, (- xp + le * p + d[i, j] + lf)/lf, 1, x_val, sgn=-1, eta=True, a=le * p2 + d[i, j])
            r3 = rg.RegionBase(0, (xp + lf * q1 - d[i, j] - lf)/le, (- xp + le * p + d[i, j] + lf)/lf, q1, x_val, sgn=-1, eta=True, a=(1 - q1) * lf + d[i,j])
            R = rg.Region((r1, r2, r3))
        elif (i, j) == (1, 0):
            r1 = rg.RegionBase((- xp + d[i, j] + le)/le, 1, 0, (xp + le * p - d[i, j] - le)/lf, x_val)
            r2 = rg.RegionBase((- xp + d[i, j] + le)/le, p1, 0, (xp + le * p - d[i, j] - le)/lf, x_val, sgn=-1, eta=True, a=(1 - p1) * le + d[i,j], bd=(True, False))
            r3 = rg.RegionBase((- xp + lf * q2 + d[i, j] + le)/le, 1, q2, (xp + le * p - d[i, j] - le)/lf, x_val, sgn=-1, eta=True, a=lf * q2 + d[i, j])
            R = rg.Region((r1, r2, r3))
        else:
            r1 = rg.RegionBase((- xp + d[i, j] + le)/le, 1, (- xp - le * p + d[i, j] + le + lf)/lf, 1, x_val)
            r2 = rg.RegionBase((- xp + d[i, j] + le)/le, p2, (- xp - le * p + d[i, j] + le + lf)/lf, 1, x_val, sgn=-1, eta=True, a=(1 - p2) * le + d[i,j], bd=(True, False))
            r3 = rg.RegionBase((- xp - lf * q2 + d[i, j] + le + lf)/le, 1, (- xp - le * p + d[i, j] + le + lf)/lf, q2, x_val, sgn=-1, eta=True, a=(1 - q2) * lf + d[i,j])
            R = rg.Region((r1, r2, r3))
        R.a_val = a
        R.b_val = b
        if b > g.d_max:
            g.d_max = b
        return R

# get L
def get_L(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val):
    if e == f:
        if i == 0:
            a = 0
        else:
            a = d
        b = (d + le)/2
        if x_val is None:
            x_val = b
        xp = theta(a, b, x_val)
        if (i, j) == (0, 0):
            r = rg.RegionBase(xp/le, 1, None, None, x_val)
        elif (i, j) == (0, 1):
            r = rg.RegionBase(0, 1 - xp/le, None, None, x_val)
        elif (i, j) == (1, 0):
            r = rg.RegionBase(1 - (xp - d)/le, 1, None, None, x_val)
        elif (i, j) == (1, 1):
            r = rg.RegionBase(0, (xp - d)/le, None, None, x_val)
        return r
    else:
        a = d[i, j]
        b = (le + lf + 
                min(d[0, 0] + d[1, 1], 
                    d[0, 1] + d[1, 0])) / 2
        if x_val is None:
            x_val = b
        xp = theta(a, b, x_val)
        if (i, j) == (0, 0):
            r = rg.RegionBase(
                    mmax(0,(2 * xp - d[0, 0] - d[0, 1] - lf)/(2 * le)), 
                    mmin(p1, (xp - d[i, j])/le), 
                    #Max(0,(2 * xp - d[0, 0] - d[0, 1] - lf)/(2 * le)), 
                    #Min(p1, (xp - d[i, j])/le), 
                    None, None, x_val, bd=(True, False))
        elif (i, j) == (0, 1):
            r = rg.RegionBase(
                    mmax(0, (2 * xp - d[0, 0] - d[0, 1] - lf)/(2 * le)), 
                    mmin(p2, (xp - d[i, j])/le), 
                    #Max(0, (2 * xp - d[0, 0] - d[0, 1] - lf)/(2 * le)), 
                    #Min(p2, (xp - d[i, j])/le), 
                    None, None, x_val, bd=(True, False))
        elif (i, j) == (1, 0):
            r = rg.RegionBase(
                    mmax(p1, (- xp + d[i, j] + le)/le), 
                    mmin(1, (- 2 * xp + d[1, 0] + d[1, 1] + 2* le + lf)/(2 * le)), 
                    #Max(p1, (- xp + d[i, j] + le)/le), 
                    #Min(1, (- 2 * xp + d[1, 0] + d[1, 1] + 2* le + lf)/(2 * le)), 
                    None, None, x_val)
        elif (i, j) == (1, 1):
            r = rg.RegionBase(
                    mmax(p2, (- xp + d[i, j] + le)/le), 
                    mmin(1, (- 2 * xp + d[1, 0] + d[1, 1] + 2* le + lf)/(2 * le)),
                    #Max(p2, (- xp + d[i, j] + le)/le), 
                    #Min(1, (- 2 * xp + d[1, 0] + d[1, 1] + 2* le + lf)/(2 * le)),
                    None, None, x_val)
        return r


def get_q(g, e, f, i, j, le, lf, d, p1, p2, q1, q2, x_val):
    x = x_val
    if e == f:
        if (i, j) == (0, 0):
            q_func = p - x / le
        elif (i, j) == (0, 1):
            q_func = p + x / le
        elif (i, j) == (1, 0):
            q_func = p - 1 + (x - d) / le
        else:
            q_func = p + 1 - (x - d) / le
        return q_func
    else:
        if (i, j) == (0, 0):
            q_func = (x - le * p - d[i, j]) / lf
        elif (i, j) == (0, 1):
            q_func = (- x + le * p + d[i, j] + lf) / lf
        elif (i, j) == (1, 0):
            q_func = (x + le * p - d[i, j] - le) / lf
        else:
            q_func = (- x - le * p + d[i, j] + le + lf) / lf
        return q_func



from enum import Enum

class Stats(Enum):
    MOMENT = 1
    CDF = 2
    PDF = 3
    CMOMENT = 4
    CCDF = 5
    CPDF = 6
#    SIMULATION = 7
#    TIMING = 8
#    SAVE = 9

    @classmethod
    def is_member(cls, value):
        return value in cls



from sympy import *
from . import commonFuncs as cf
from sympy.abc import p, q, x
from . import numericFuncs as nf

# basic components of a region
class RegionBase:
    # (pl, pu, ql, qu) tuple representation,
    # sgn: remove or add;
    # eta: a gate function to control the support
    # a, b: the eta function bounds
    # bd: indicate whether the corresponding boundary 
    #     (just pl, pu for now) is included
    #       in the region
    def __init__(self, pl, pu, ql, qu, xval, sgn=1, eta=False, a=0, b=oo, bd=(True, True)):
        self.pl = sympify(pl)
        self.pu = sympify(pu)
        if ql is not None:
            self.ql = sympify(ql)
        else:
            self.ql = None
        if qu is not None:
            self.qu = sympify(qu)
        else:
            self.qu = None
        self.sgn = sgn
        self.eta = eta
        self.a = a
        self.b = b
        self.xval = xval # is x if it is a region function; or upper bound if is a region
        self.bd = bd
        # self.modules = ['numpy']

    # the mass function
    def m(self, f):
        if self.ql is None:
            res = integrate(f, (p, self.pl, self.pu))
        else:
            res = integrate(f, (q, self.ql, self.qu), (p, self.pl, self.pu))
        return res

    # adjusted mass function
    def m_adj(self, f):
        res = self.m(f)
        if self.eta is True:
            return self.sgn * cf.eta(self.a, self.b, self.xval) * res
        else:
            return self.sgn * res

    # conditional mass function
    def m_p(self, f):
        myeta = None
        if self.bd[0] is False:
            myeta = cf.etal
        elif self.bd[1] is False:
            myeta = cf.etar
        else:
            myeta = cf.eta

        return myeta(self.pl, self.pu, p) * integrate(f, (q, self.ql, self.qu))

    # conditional adjusted mass function
    def m_p_adj(self, f):
        res = self.m_p(f)
        if self.eta is True:
            return self.sgn * cf.eta(self.a, self.b, self.xval) * res
        else:
            return self.sgn * res

    # numerical integration
    def m_num(self, expr, x_val=None):
        if x_val is None:
            range1 =  (p, self.pl, self.pu)
        else:
            range1 = (p, self.pl.subs(x, x_val), self.pu.subs(x, x_val))

        if self.ql is None:
            range2 = None
        elif x_val is None:
            range2 = (q, self.ql, self.qu)
        else:
            range2 = (q, self.ql.subs(x, x_val), self.qu.subs(x, x_val))

        m, e = nf.N_int(expr, range1, range2)
        # print('expr\t', expr)
        # print('range1\t', range1)
        # print('range2\t', range2)
        return m

    # adjusted numerical integration
    def m_num_adj(self, expr, x_val=None):
        res = self.m_num(expr, x_val)
        return self.sgn * nf.eta(self.a, self.b, self.xval) * res

    # produce fixed region bases
    def RB_const(self, val):
        return RegionBase(
                #self.pl,
                #self.pu,
                #self.ql,
                #self.qu,
                self.eval(self.pl, val),
                self.eval(self.pu, val),
                self.eval(self.ql, val),
                self.eval(self.qu, val),
                sgn = self.sgn,
                eta = self.eta,
                a = self.a,
                b = self.b,
                xval = val,
                bd = self.bd
                )

    # def lambdify
    def eval(self, expr, x_val):
        return expr.subs(x, x_val)

    # print region description
    def print(self):
        print('pl:\t', self.pl)
        print('pu:\t', self.pu)
        print('ql:\t', self.ql)
        print('qu:\t', self.qu)
        print('a:\t', self.a)
        print('b:\t', self.b)
        print('xval:\t', self.xval)


# a list of region bases
class Region:
    def __init__(self, bases):
        self.bases = bases
   
    # the measure of this region
    def m(self, f):
        res = 0
        for b in self.bases:
            res = res + b.m_adj(f)
        return res

    # the conditional measure of this region
    def m_p(self, f, p_val=None):
        res = 0
        for b in self.bases:
            res = res + b.m_p_adj(f)
        return res

    # numerical value
    def m_num(self, f, x_val=None):
        res = 0
        for b in self.bases:
            res += b.m_num_adj(f, x_val)
        return res

    # print region description
    def print(self):
        for i, b in enumerate(self.bases):
            print('Piece ',i+1, ':\n')
            b.print()
            print('\n')

    # generate constant Regions 
    def R_const(self, val):
        rs = []
        for b in self.bases:
            rs.append(b.RB_const(val))
        return Region(rs)

# check whether the input is a number or not
def isnumber(x):
    tstr = str(type(x))
    return 'numbers' in tstr or 'int' in tstr or 'float' in tstr

if __name__ == '__main__':
    le = 3
    mi = Function('min')
    r1 = RegionBase(0, x/le, 0, p)
    r2 = RegionBase(x/le, mi(1,3), p-x/le, p, eta=True, a=1, b=3)

    f = 2 * p

#    print(r1.m(f))
#    print(r2.m(f))
    
    R = Region((r1,r2))
    R.assign_f(f)
    print(R.m)
    R.m_lambdify()
    print(R.m_func(0))
    print(R.m_func(0.9))
    print(R.m_func(1))
    print(R.m_func(2))
    print(R.m_func(3))
    print(R.m_func(3.1))
    print(R.m_func(4))




from .modules.enums import Stats
from .modules.basicInfo import readGraph
from .modules.numericFuncs import pdf_check
from .modules.commonFuncs import load_formulas, get_largest_component
from .modules.phis import Phi
from .modules.dataCollector import data_collector
from .modules.formulas import Formulas







from scipy.integrate import quad, dblquad
from . import numericFuncs as nf
import math

# basic components of a region
class RegionBase:
    def __init__(self, pl, pu, ql, qu, xval, bd=(True, True)):
        self.pl = pl
        self.pu = pu
        self.ql = ql
        self.qu = qu
        self.xval = xval # is x if it is a region function; or upper bound if is a region
        self.bd = bd
        # self.modules = ['numpy']

    # the mass function
    def m(self, func):
        if self.ql is None:
            region = (self.pl, self.pu)
            int_func = quad
        else:
            int_func = dblquad
            region = (self.pl, self.pu, self.ql, self.qu)

        return int_func(func, *region)[0]

    # conditional mass function
    def m_p(self, f, p_val):
        myeta = None
        if self.bd[0] is False:
            myeta = nf.etal
        elif self.bd[1] is False:
            myeta = nf.etar
        else:
            myeta = nf.eta

        etaval = myeta(self.pl, self.pu, p_val)
        if etaval != 0:
            return quad(f, self.ql(p_val), self.qu(p_val))[0]
        else:
            return 0

    # print region description
    def print(self):
        print('pl:\t', self.pl)
        print('pu:\t', self.pu)
        print('ql:\t', self.ql)
        print('qu:\t', self.qu)
        print('xval:\t', self.xval)




import os
import errno


# function write file
# help create folders recursively if doesn't exist
# support append as well as write
def wFile(ffname, output, writetype='a'):
    dirname = os.path.dirname(ffname)
    if not os.path.exists(dirname):
        try:
            os.makedirs(dirname)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise
    if not os.path.isfile(ffname):
        wtype = 'w'
    else:
        wtype = writetype

    wf = open(ffname, wtype)
    wf.write(output)
    wf.close()



from .enums import Stats
from . import numericFuncs as nf
from sympy import *
from sympy.abc import p, q, x
from .plot import plot1d

# serialized formula
class SFormula:
    def __init__(self):
        self.stat = None
        self.fs = {}
        self.N_fs = {}
        self.mods = None
        self.idx_num = 0  # dimension of indexes for self.fs
        self.val_keys = None  # the symbols for values
        self.plot_info = (None, x, 0, 0)  # plotting parameters
        self.unikey = None

    def eval(self, *params):
        # split params into keys and vals
        keys, subval, vals = self.__split_input(*params)
        if keys == self.unikey:
            unikey = (self.unikey,)
            if unikey in self.N_fs:
                N_f = self.N_fs[unikey]
            else:
                print('This function does not exist!')
                return False
        else:
            if keys in self.N_fs:
                N_f = self.N_fs[keys]
            else:
                print('This function does not exist!')
                return False

        return self.__apply(N_f, vals)

    # output the formula
    def formula(self, *params):
        # split params into keys and vals
        keys, subval, vals = self.__split_input(*params)
        if keys == self.unikey:
            unikey = (self.unikey,)
            if unikey in self.fs:
                myf = self.fs[unikey]
                return self.__subs(myf, subval)
            else:
                print('This function does not exist!')
                return False
        else:
            if keys in self.fs:
                myf = self.fs[keys]
                return self.__subs(myf, subval)
            else:
                print('This function does not exist!')
                return False

    # input f_info = (f, var, lb, ub)
    def plot(self, *params, step=0.01, show=True):
        if self.stat != Stats.MOMENT:
            unikey = (self.unikey,)
            if len(params) == 0:
                if unikey in self.fs:
                    f = self.fs[unikey]
                else:
                    print('This function does not exist!')
                    return False
            else:
                # split params into keys and vals
                keys, subval, vals = self.__split_input(*params)
                keys = self.mkkey(keys)
                if keys in self.fs:
                    f = self.__subs(self.fs[keys], subval)
                else:
                    print('This function does not exist!')
                    return False

            var = self.plot_info[1]
            lb = self.plot_info[2]
            ub = self.plot_info[3]

            f_lambda = lambdify(var, f, modules=self.mods)
            plot1d(f_lambda, lb, ub, step, show=True)

        else:
            print('moment function cannot be plotted.')

    def __apply(self, f, vals):
        if len(vals) == 0:
            return f
        else:
            return f(*vals)

    def mkkey(self, keys):
        return keys

    def __split_input(self, *params):
        if self.idx_num == 0:
            keys = self.unikey
            vals = params
        else:
            keys = []
            vals = []
            for i, v in enumerate(params):
                if i < self.idx_num:
                    keys.append(v)
                else:
                    vals.append(v)
            keys = tuple(keys)
        
        # make subs info
        if len(vals) == 0:
            subval = None
        elif len(vals) == 1:
            subval = [(self.val_keys[0], vals[0])]
        else:
            subval = list(zip(self.val_keys, vals))

        vals = tuple(vals)

        return (keys, subval, vals)

    # return formula with subs
    def __subs(self, formula, subval):
        if subval is None:
            return formula
        else:
            return formula.subs(subval)

